import tiktoken
from langchain.output_parsers import ResponseSchema, StructuredOutputParser
from langchain.prompts import PromptTemplate

from goose_checker.context import TerraformContext, get_terraform_context
from goose_checker.diff import GitDiff
from goose_checker.models import GooseChecker, GooseCheckerResponse
from goose_checker.prompts import (aggregation_prompt, base_prompt,
                                   terraform_prompt)
from goose_checker.response_schema import (approve_or_deny_schema,
                                           other_issues_explanation_schema,
                                           silly_mistakes_explaination_schema,
                                           silly_mistakes_aggregation_schema)


def get_token_len(string, model_name: str = "gpt-3.5-turbo"):
    encoding = tiktoken.encoding_for_model(model_name)  # Make this dynamic
    num_tokens = len(encoding.encode(string))
    return num_tokens


def _build_context_subprompt(context: TerraformContext, quota: int):
    subprompt = f"""
    file_name: {context.file_name}

    contents: {context.contents}
    """

    # TODO: Probably a more intelligent way to do this...
    if get_token_len(subprompt) > quota:
        proportion = quota / get_token_len(subprompt)
        subprompt = subprompt[: int(proportion * len(subprompt))]

    return subprompt


def _build_diff_subprompt(diff: GitDiff, quota: int) -> dict:
    subprompt = f"""
    file_name: {diff.file_name}
    diff: {diff.diff}
    """
    if get_token_len(subprompt) > quota:
        proportion = quota / get_token_len(subprompt)
        subprompt = subprompt[: int(proportion * len(subprompt))]

    tokens_used = get_token_len(subprompt)

    return {"subprompt": subprompt, "tokens_used": tokens_used}


def base_checker(diff: GitDiff, goose_checker: GooseChecker) -> GooseCheckerResponse:

    # Core chain to analyze terraform diff
    response_schemas = [
        silly_mistakes_explaination_schema,
        other_issues_explanation_schema,
    ]
    output_parser = StructuredOutputParser.from_response_schemas(response_schemas)
    format_instructions = output_parser.get_format_instructions()
    model = goose_checker._get_model()

    diff_and_tokens_used: dict = _build_diff_subprompt(diff, goose_checker.quota)

    prompt_template = PromptTemplate.from_template(
        base_prompt,
        partial_variables={
            "format_instructions": format_instructions,
            "diff_subprompt": diff_and_tokens_used["subprompt"],
        },
    )

    chain = prompt_template | model | output_parser

    output = chain.invoke({})
    return GooseCheckerResponse(
        file_name=diff.file_name,
        silly_mistakes=output["silly_mistakes_explanation"],
        other_issues=output["other_issues_explanation"],
    )


def terraform_checker(
    diff: GitDiff, goose_checker: GooseChecker
) -> GooseCheckerResponse:

    # Core chain to analyze terraform diff
    response_schemas = [
        silly_mistakes_explaination_schema,
        other_issues_explanation_schema,
    ]
    output_parser = StructuredOutputParser.from_response_schemas(response_schemas)
    format_instructions = output_parser.get_format_instructions()
    model = goose_checker._get_model()
    format_len = get_token_len(format_instructions)

    # 1. Build Context
    terraform_context = get_terraform_context(diff)
    # 2. get our model
    model = goose_checker._get_model()
    # 3. Build our Chain using context
    diff_and_tokens_used: dict = _build_diff_subprompt(diff, goose_checker.quota)

    quota_remain = (
        goose_checker.quota - diff_and_tokens_used["tokens_used"] - format_len
    )

    context_subprompt = ""
    for context in terraform_context:

        context_sub_subprompt = _build_context_subprompt(context, quota_remain)
        context_subprompt += context_sub_subprompt
        quota_remain = quota_remain - get_token_len(context_sub_subprompt)

        if quota_remain < 250:
            break

    prompt_template = PromptTemplate.from_template(
        terraform_prompt,
        partial_variables={
            "format_instructions": format_instructions,
            "context_subprompt": context_subprompt,
            "diff_subprompt": diff_and_tokens_used["subprompt"],
        },
    )

    chain = prompt_template | model | output_parser

    output = chain.invoke({})
    return GooseCheckerResponse(
        file_name=diff.file_name,
        silly_mistakes=output["silly_mistakes_explanation"],
        other_issues=output["other_issues_explanation"],
    )


def approve_or_deny(
    responses: list[GooseCheckerResponse], goose_checker: GooseChecker
) -> tuple[bool, str]:
    response_schemas = [approve_or_deny_schema, silly_mistakes_aggregation_schema]
    output_parser = StructuredOutputParser.from_response_schemas(response_schemas)
    format_instructions = output_parser.get_format_instructions()
    model = goose_checker._get_model()

    evaluations_subprompt = ""

    for response in responses:
        evaluations_subprompt = f"""
        file_name: {response.file_name}
        silly_mistakes: {response.silly_mistakes}
        """

    prompt_template = PromptTemplate.from_template(
        aggregation_prompt,
        partial_variables={
            "format_instructions": format_instructions,
            "evaluations_subprompt": evaluations_subprompt,
        },
    )

    chain = prompt_template | model | output_parser

    result = chain.invoke({})

    return (result["approve_or_deny"], result["silly_mistakes"])



import argparse
from concurrent.futures import ThreadPoolExecutor, as_completed
import os

from dotenv import load_dotenv

from goose_checker.chains import (base_checker, approve_or_deny,
                                  terraform_checker)
from goose_checker.diff import get_git_diffs
from goose_checker.goose_ascii import detected_ascii, goose_ascii
from goose_checker.models import (AzureGooseChecker, GooseCheckerResponse,
                                  OpenAIGooseChecker)


def run_checker(diff, goose_checker):
    if diff.file_name.endswith(".tf"):
        return terraform_checker(diff=diff, goose_checker=goose_checker)
    else:
        return base_checker(diff=diff, goose_checker=goose_checker)


def display_issues(responses: list[GooseCheckerResponse]):
    for response in responses:
        if response.silly_mistakes:  # Checks if silly_mistakes is not empty
            print("\n")
            print(f"File Name: {response.file_name}")
            print(f"Silly Mistakes: {response.silly_mistakes}")
            print(f"Other Issues: {response.other_issues}")
            print("\n")


def cli_main():
    parser = argparse.ArgumentParser(description="Checks if you goofed")
    parser.add_argument(
        "--branch",
        nargs="?",
        default=os.environ.get("GOOSE_CHECKER_BRANCH_NAME", "main"),
        help="The branch or commit to compare against",
    )
    parser.add_argument(
        "--model",
        nargs="?",
        default=os.environ.get("GOOSE_CHECKER_MODEL_NAME", "gpt-3.5-turbo"),
        help="LLM model to use, for Azure instances this is the deployment ID.",
    )

    supported_providers = ["azure", "openai"]
    parser.add_argument(
        "--provider",
        nargs="?",
        default=os.environ.get("GOOSE_CHECKER_PROVIDER","openai"),
        choices=supported_providers,
        help=f"The cloud provider to use, choices are: {supported_providers}",
    )

    args = parser.parse_args()

    print(f"Checking for geese by comparing to {args.branch}")
    print(f"Using model: {args.model} provided by {args.provider}")

    if args.provider == "azure":
        goose_checker = AzureGooseChecker(
            with_respect_to=args.branch, deployment_name=args.model
        )
    elif args.provider == "openai":
        goose_checker = OpenAIGooseChecker(
            with_respect_to=args.branch, model=args.model
        )
    else:
        raise ValueError(f"Provider specified not supported: {args.provider}")

    # Lets Cook

    diffs = get_git_diffs(with_respect_to=args.branch)

    all_responses = []
    # Initialize ThreadPoolExecutor
    with ThreadPoolExecutor(max_workers=4) as executor:
        # Submit tasks to the executor
        future_to_diff = {
            executor.submit(run_checker, diff, goose_checker): diff for diff in diffs
        }
        # Retrieve and store results as they complete
        for future in as_completed(future_to_diff):
            diff = future_to_diff[future]
            try:
                resp = future.result()
                all_responses.append(resp)
            except Exception as exc:
                print(f"Failed to analyze {diff.file_name} due to {exc}")
    if not all_responses:
        return False

    approved, explanation = approve_or_deny(all_responses, goose_checker)

    if not approved:
        print(goose_ascii)
        print(detected_ascii)
        print(explanation)
    
    else:
        print("Approved! No Geese detected.")

    return approved


if __name__ == "__main__":
    load_dotenv()
    cli_main()



import glob
from pathlib import Path

from pydantic import BaseModel

from goose_checker.diff import GitDiff


class BaseContext(BaseModel):
    file_name: str
    contents: str


class TerraformContext(BaseContext):
    pass


def get_terraform_context(diff: GitDiff) -> list[TerraformContext]:

    ret = []

    file_path = Path(diff.file_name)
    # Use glob to find matching files excluding the current file's directory
    pattern = f"**/{file_path.name}"
    for entry in glob.glob(pattern, recursive=True):
        path = Path(entry)
        if path != file_path and path.name == file_path.name:
            with open(path, "r") as f:
                contents = f.read()
                ret.append(TerraformContext(file_name=path.name, contents=contents))
    return ret



import os

from langchain_openai import AzureChatOpenAI, ChatOpenAI
from pydantic import BaseModel


class GooseChecker(BaseModel):
    with_respect_to: str  # The branch or commit to compare against

    @property
    def quota(self) -> int:
        return 4095

    def _get_model(self):
        raise NotImplementedError("Subclasses must implement this method")


class GooseCheckerResponse(BaseModel):
    file_name: str
    silly_mistakes: str
    other_issues: str


class OpenAIGooseChecker(GooseChecker):
    model: str

    @property
    def quota(self) -> int:
        if self.model == "gpt-3.5-turbo":
            return 4095
        elif self.model == "gpt-4":
            return 8190

        return 4095

    @property
    def api_key(self) -> str:
        try:
            return os.environ["OPENAI_API_KEY"]
        except KeyError:
            raise ValueError("Environment variable OPENAI_API_KEY is not set")

    def _get_model(self):
        return ChatOpenAI(model=self.model, api_key=self.api_key, temperature=0.1)


class AzureGooseChecker(GooseChecker):
    deployment_name: str

    @property
    def quota(self) -> int:
        if "gpt-3.5-turbo" in self.deployment_name:
            return 4095
        elif "gpt-4" in self.deployment_name:
            return 8190

        return 4095

    @property
    def openai_api_version(self) -> str:
        try:
            return os.environ["AZURE_OPENAI_API_VERSION"]
        except KeyError:
            raise ValueError("Environment variable AZURE_OPENAI_API_VERSION is not set")

    @property
    def api_key(self) -> str:
        try:
            return os.environ["AZURE_OPENAI_API_KEY"]
        except KeyError:
            raise ValueError("Environment variable AZURE_OPENAI_API_KEY is not set")

    @property
    def base_url(self) -> str:
        try:
            return os.environ["AZURE_OPENAI_BASE_URL"]
        except KeyError:
            raise ValueError("Environment variable AZURE_OPENAI_BASE_URL is not set")

    def _get_model(self):
        return AzureChatOpenAI(
            deployment_name=self.deployment_name,  # Want to try another model? Check out the AI-Playground in the Analytics Lab Launcher for examples using other models.
            api_version=self.openai_api_version,
            azure_endpoint=self.base_url,
            temperature=0.1,  # a number from 0 to 2
        )



from langchain.output_parsers import ResponseSchema

approve_or_deny_schema = ResponseSchema(
    name="approve_or_deny",
    description="A boolean True or False to approve or deny.",
)
silly_mistakes_aggregation_schema = ResponseSchema(
    name="silly_mistakes",
    description="If there are any silly mistakes, explain clearly how to fix them here. Use numbered steps if possible.",
)

other_issues_explanation_schema = ResponseSchema(
    name="other_issues_explanation",
    description="A brief explanation. This is where you can include any other issues you find relating to performance or improving the code. If there are no other issues, you can leave this field blank.",
)

silly_mistakes_explaination_schema = ResponseSchema(
    name="silly_mistakes_explanation",
    description="A brief explanation of any silly mistakes found in the code. If there are no silly mistakes, you can leave this field blank.",
)



terraform_prompt = """
<< INSTRUCTIONS >>
You are an software engineer reviewing a pull request by analyzing the DIFF of a terraform file
Consider the following code snippets provided as CONTEXT, keep in mind they may not be complete


<< CONTEXT >>
{context_subprompt}

<< DIFF >>
{diff_subprompt}

<< TASK >>

Focus on the following:
1. Naming conventions and idiomatic code
2. Consistency between CONTEXT and the DIFF

<< FORMAT INSTRUCTIONS >>
{format_instructions}
"""

base_prompt = """
<< INSTRUCTIONS >>
You are an software engineer reviewing a pull request by analyzing the DIFF
Keep in mind the DIFF may not be complete

<< DIFF >>
{diff_subprompt}

<< TASK >>

Focus on the following:
1. Naming conventions and idiomatic code
2. Clean Code


<< FORMAT INSTRUCTIONS >>
{format_instructions}

"""


aggregation_prompt = """
<< INSTRUCTIONS >>

You are a manager determining the whether code should be accepted or denied.
You will be given a list of EVALUATIONS done by your teams engineers.
Your job is to APPROVE or DENY the code by looking at the review.

<< EVALUATIONS >>

{evaluations_subprompt}

<< FORMAT INSTRUCTIONS >>

{format_instructions}
"""
